"""
Put the content from the local directory and put it in the remote bucket.
"""

from widely.bucket import get_current_bucket
from widely.diff import generate_diffs, show_diffs, run_diffs
from widely.util import get_y_or_n


def push():
    """
    Pushes local content to AWS S3 for publication.

    Usage: widely push
    """
    bucket = get_current_bucket()
    diffs = generate_diffs(bucket)
    if diffs:
        show_diffs(diffs)
        # Ask for approval
        decision = get_y_or_n(
            'Would you like to make the changes in {0}?'.format(
                bucket.name
            )
        )
        if decision:
            run_diffs(diffs, bucket, local_changes=False)
    else:
        print('There are no changes to push.')



from widely.bucket import get_current_or_specified_bucket


def domains(arguments):
    """
    Displays a list of domains for the current or specified sitename.

    Usage: widely domains [--site <SITENAME>]
    """
    bucket = get_current_or_specified_bucket(arguments)
    print('=== {0} Domain Names'.format(bucket.name))
    print(bucket.get_website_endpoint())



"""
Helper functions not widely specific.
"""

from __future__ import print_function


YES_RESPONSES = set(['y', 'Y', 'Yes', 'yes'])
NO_RESPONSES = set(['n', 'N', 'No', 'no'])


def get_y_or_n(message=None, input=raw_input, output=print):
    """
    Asks the user whether to proceed.
    """
    if not message:
        message = "Continue? (y/n)"
    while True:
        response = input(message + ' ')
        if response in YES_RESPONSES:
            return True
        elif response in NO_RESPONSES:
            return False
        output('Please enter y/n.')


def sizeof_fmt(num):
    """
    Formats number of bytes into a human readable form.

    From:
    """
    for size in ['bytes', 'KB', 'MB', 'GB']:
        if -1024.0 < num < 1024.0:
            return "%3.1f %s" % (num, size)
        num /= 1024.0
    return "%3.1f %s" % (num, 'TB')



"""
Get the content from the specified bucket and put it in the local directory.
"""

from widely.bucket import get_current_or_specified_bucket, \
    get_current_or_specified_sitename
from widely.diff import generate_diffs, show_diffs, run_diffs
from widely.util import get_y_or_n


def pull(arguments):
    """
    Pulls content associated with the specified sitename from AWS S3
    to the current directory.

    Usage: widely pull --site <SITENAME>
    """
    sitename = get_current_or_specified_sitename(arguments)
    bucket = get_current_or_specified_bucket(arguments)
    diffs = generate_diffs(bucket)
    if diffs:
        show_diffs(diffs)
        # Ask for approval
        if get_y_or_n("Would you like to make the changes locally?"):
            run_diffs(diffs, bucket, local_changes=True)
            # Set the .widely file (or create it) with the bucket name
            with open('.widely', 'w') as f:
                f.write(sitename)
    else:
        print('There are no changes to pull.')



"""
Get the status of various S3 regions.
"""

import feedparser
from prettytable import PrettyTable


def status():
    """
    Displays AWS S3 status.

    Usage: widely status
    """
    s3_status_rss_feeds = {
        "N. California": "http://status.aws.amazon.com/rss/s3-us-west-1.rss",
        "Oregon": "http://status.aws.amazon.com/rss/s3-us-west-2.rss",
        "US Standard": "http://status.aws.amazon.com/rss/s3-us-standard.rss",
        "Sao Paulo": "http://status.aws.amazon.com/rss/s3-sa-east-1.rss",
        "Ireland": "http://status.aws.amazon.com/rss/s3-eu-west-1.rss",
        "Singapore": "http://status.aws.amazon.com/rss/s3-ap-southeast-1.rss",
        "Sydney": "http://status.aws.amazon.com/rss/s3-ap-southeast-2.rss",
        "Tokyo": "http://status.aws.amazon.com/rss/s3-ap-northeast-1.rss"
    }
    print('=== S3 Status')
    table = PrettyTable(['Region', 'Status'])
    for region_name, rss_url in s3_status_rss_feeds.iteritems():
        feed = feedparser.parse(rss_url)
        if feed.entries:
            region_status = feed.entries[0].title
        else:
            region_status = 'Unknown.'
        table.add_row([region_name, region_status])
    print(table)



"""
Helper methods for working with S3 buckets.
"""

import sys

from boto.s3.connection import S3Connection
from boto.exception import S3ResponseError

from widely.util import sizeof_fmt


class NoSuchBucket(Exception):
    """
    Raised when there is no bucket for the specified sitename.
    """
    pass


class NoWidelyDotfile(Exception):
    """
    Raised when there is no .widely in the current directory.
    """
    pass


def get_buckets():
    """
    Returns a list of all accessible buckets.
    """
    from widely.commands.auth import get_credentials
    get_credentials()
    conn = S3Connection()
    buckets = conn.get_all_buckets()
    return buckets


def get_specified_bucket(sitename):
    """
    Returns the bucket for the specified sitename.
    """
    from widely.commands.auth import get_credentials
    get_credentials()
    conn = S3Connection()
    try:
        bucket = conn.get_bucket(sitename)
        bucket.get_website_configuration()
    except S3ResponseError:
        raise NoSuchBucket
    return bucket


def get_current_bucket():
    """
    Returns the bucket associated with the current directory, if it is
    a widely directory.
    """
    try:
        sitename = open('.widely', 'r').read().split()[0]
    except IOError:
        raise NoWidelyDotfile
    return get_specified_bucket(sitename)


def get_current_or_specified_bucket(arguments):
    """
    Returns a bucket. If one was specified and it exists, it is that
    one, otherwise, it is the current directory's bucket if it exists.
    """
    sitename = get_current_or_specified_sitename(arguments)
    try:
        return get_specified_bucket(sitename)
    except NoSuchBucket:
        print(' !\tSite not found')
        sys.exit(1)


def websites_from_buckets(buckets):
    """
    Filter a list of buckets into only those which are configured to
    be a website.
    """
    for bucket in buckets:
        try:
            bucket.get_website_configuration()
            yield bucket
        except S3ResponseError:
            continue


def bucket_size(bucket):
    """
    Returns the total number of bytes in a bucket.
    """
    return sum(key.size for key in bucket.get_all_keys())


def readable_bucket_size(bucket):
    """
    Returns the number of bytes in a bucket in a human readable form.
    """
    return sizeof_fmt(bucket_size(bucket))


def get_current_or_specified_sitename(arguments):
    """
    Returns the bucket name/sitename from the arguments if there was
    one specified, otherwise from the .widely if it exists.
    """
    sitename = arguments['<SITENAME>']
    if sitename:
        return sitename
    else:
        try:
            with open('.widely', 'r') as f:
                return f.read().split()[0]
        except IOError:
            print(' !\tNo site specified.')
            print(" !\tRun this command from a site folder or specify which "
                  "site to use with --site SITENAME.")
            sys.exit(1)



#! /usr/bin/env python
"""
Widely.

Usage:
  widely (help | -h | --help) [<TOPIC>]
  widely [login | auth:login]
  widely [logout | auth:logout]
  widely auth:whoami
  widely sites
  widely sites:info [--site <SITENAME>]
  widely sites:create <SITENAME>
  widely sites:copy <SITENAME>
  widely sites:rename <SITENAME>
  widely sites:delete <SITENAME>
  widely domains [--site <SITENAME>]
  widely local [-p <PORT> | --port <PORT>]
  widely push
  widely pull --site <SITENAME>
  widely open [--site <SITENAME>]
  widely status
  widely logs [--site <SITENAME>]
  widely (version | -v | --version)

Options:
  -p --port
  -h --help
  -v --version
"""

import sys
import pkg_resources

from docopt import docopt

from widely.commands.help import _help
from widely.commands.push import push
from widely.commands.pull import pull
from widely.commands.status import status
from widely.commands.sites import sites, sites_create, sites_info, \
    sites_delete, sites_rename, sites_copy
from widely.commands.local import local
from widely.commands.logs import logs
from widely.commands.version import version
from widely.commands.auth import auth_login, auth_logout, auth_whoami
from widely.commands.domains import domains
from widely.commands.open import _open


def main():
    """
    Dispatch based on arguments.

    """
    version_string = ''.join([
        'widely/',
        str(pkg_resources.require('widely')[0].version),
        ' python/',
        sys.version
    ])

    arguments = docopt(__doc__, version=version_string, help=False)
    if arguments['version']:
        version(version_string)
    elif arguments['--help'] or arguments['help']:
        _help(arguments)
    elif arguments['auth:login'] or arguments['login']:
        auth_login()
    elif arguments['auth:logout'] or arguments['logout']:
        auth_logout()
    elif arguments['auth:whoami']:
        auth_whoami()
    elif arguments['domains']:
        domains(arguments)
    elif arguments['local']:
        local(arguments)
    elif arguments['logs']:
        logs(arguments)
    elif arguments['open']:
        _open(arguments)
    elif arguments['sites']:
        sites()
    elif arguments['sites:create']:
        sites_create(arguments)
    elif arguments['sites:info']:
        sites_info(arguments)
    elif arguments['sites:rename']:
        sites_rename(arguments)
    elif arguments['sites:copy']:
        sites_copy(arguments)
    elif arguments['sites:delete']:
        sites_delete(arguments)
    elif arguments['status']:
        status()
    elif arguments['push']:
        push()
    elif arguments['pull']:
        pull(arguments)
    else:
        print('We did not recognize your specified arguments.')
        _help(arguments)
        sys.exit(2)
    sys.exit()


if __name__ == '__main__':
    main()



from widely.bucket import get_current_or_specified_bucket


def logs(arguments):
    """
    Displays access and status logs for the current site or specified
    sitename.

    Usage: widely logs [--site <SITENAME>]
    """
    bucket = get_current_or_specified_bucket(arguments)
    print('Logs for {0}'.format(bucket.name))
    # TODO:
    # - Get the logs
    # - Reformat them (in chronological order)
    # - Print them
    print('Sorry, this is not yet implemented!')



"""
Manage the authorization tokens.
"""

import os
from ConfigParser import DuplicateSectionError, NoSectionError

import boto


def auth_login():
    """
    Saves user authentication data for AWS S3 using boto. This does not
    verify your credentials.

    Usage: widely auth:login
    """
    print('Enter your AWS credentials.')
    aws_access_key_id = raw_input('Access Key ID: ')
    aws_secret_access_key = raw_input('Secret Access Key ID: ')
    try:
        boto.config.add_section('Credentials')
    except DuplicateSectionError:
        pass

    # TODO: Check for valid credentials ?
    boto.config.set(
        'Credentials',
        'aws_access_key_id',
        aws_access_key_id
    )
    boto.config.set(
        'Credentials',
        'aws_secret_access_key',
        aws_secret_access_key
    )

    boto_config_file = open(os.path.expanduser('~/.boto'), 'w')
    boto.config.write(boto_config_file)
    print('Authentication successful.')
    return aws_access_key_id, aws_secret_access_key


def auth_logout():
    """
    Clears any locally stored authentication data from boto's config
    for AWS S3.

    Usage: widely auth:logout
    """
    try:
        boto.config.remove_option('Credentials', 'aws_access_key_id')
        boto.config.remove_option('Credentials', 'aws_secret_access_key')
    except NoSectionError:
        pass
    boto_config_file = open(os.path.expanduser('~/.boto'), 'w')
    boto.config.write(boto_config_file)
    print('Local credentials cleared.')


def auth_whoami():
    """
    Displays locally stored access key.

    Usage: widely auth:whoami
    """
    aws_access_key_id, _ = get_credentials()
    print(aws_access_key_id)


def get_credentials():
    """
    Returns current access_key and secret_access_key or asks the user to
    login.
    """
    try:
        aws_access_key_id = boto.config.get(
            'Credentials',
            'aws_access_key_id'
        )
        aws_secret_access_key = boto.config.get(
            'Credentials',
            'aws_secret_access_key'
        )
        assert aws_access_key_id is not None
        assert aws_secret_access_key is not None
    except AssertionError:
        print('No credentials found.')
        return auth_login()
    return aws_access_key_id, aws_secret_access_key



"""
Work with sites, which are abstracted S3 buckets.
"""

import sys
import os.path

from boto.s3.connection import S3Connection
from boto.exception import S3CreateError

from widely.commands.push import push
from widely.bucket import get_buckets, websites_from_buckets, \
    get_current_or_specified_bucket, get_current_bucket, \
    get_specified_bucket, readable_bucket_size
from widely.util import get_y_or_n


def sites():
    """
    Displays sites, which are simply the set of buckets which are web sites.

    Usage: widely sites
    """
    buckets = get_buckets()
    websites = websites_from_buckets(buckets)
    print('=== My sites')
    for site in websites:
        print(site.name)


def sites_info(arguments):
    """
    Displays detailed information about the current or specified sitename.

    Usage: widely sites:info [--site <SITENAME>]

    """
    bucket = get_current_or_specified_bucket(arguments)
    print('=== {0}'.format(bucket.name))
    print('Owner:   {0}'.format(bucket.get_acl().owner.display_name))
    print('Size:    {0}'.format(readable_bucket_size(bucket)))
    print('Web URL: {0}'.format(bucket.get_website_endpoint()))


def sites_create(arguments):
    """
    Creates a new site for the specified sitename.

    Usage: widely sites:create <SITENAME>
    """
    # TODO: Randomly assigned site names.
    # make sure there is no .widely file locally

    if os.path.exists('.widely'):
        print('This directory is already a widely site.')
        sys.exit(1)
    from widely.commands.auth import get_credentials
    get_credentials()
    conn = S3Connection()
    sitename = arguments['<SITENAME>']
    try:
        bucket = conn.create_bucket(sitename)
        assert bucket.get_all_keys() == []
        # Set bucket configuration
        error_key = raw_input(
            'Please enter the key of your 404 page (default: 404.html): ')
        if not error_key:
            error_key = '404.html'
        bucket.configure_website(
            suffix='index.html',
            error_key=error_key
        )
        # Configure policy
        # TODO: Ask user first?
        bucket.set_acl('public-read')
        # TODO: Configure logging
        with open('.widely', 'w') as f:
            f.write(sitename)
        return push()
    except S3CreateError:
        print('A site with that name already exists.')
        sys.exit()
    except AssertionError:
        print('This bucket already has keys.')
        sys.exit()


def sites_copy(arguments):
    """
    Copies the site in the current directory to the new sitename.

    Usage: widely sites:copy <SITENAME>
    """
    current_bucket = get_current_bucket()
    new_bucket_name = arguments['<SITENAME>']
    try:
        assert current_bucket.name != new_bucket_name
    except AssertionError:
        print('Cannot rename current bucket to current bucket.')
        sys.exit()
    sites_create(arguments)
    new_bucket = get_specified_bucket(new_bucket_name)
    for key in current_bucket.get_all_keys():
        new_bucket.copy_key(key.name, current_bucket.name, key.name)
    print('{0} copied to {1}'.format(current_bucket.name, new_bucket_name))


def sites_rename(arguments):
    """
    Renames the site in the current directory to the new sitename.

    Usage: widely sites:rename <SITENAME>
    """
    sites_copy(arguments)
    new_sitename = arguments['<SITENAME>']
    bucket = get_current_bucket()

    decision = get_y_or_n(
        'Would you like to delete {0}?'.format(bucket.name)
    )

    if decision:
        # Update the .widely to the new sitename
        with open('.widely', 'w') as f:
            f.write(new_sitename)
        bucket.delete_keys(bucket.get_all_keys())
        bucket.delete()


def sites_delete(arguments):
    """
    Deletes specified bucket from AWS S3.

    Usage: widely sites:delete <SITENAME>
    """
    bucket = get_specified_bucket(arguments['<SITENAME>'])
    bucket.delete_keys(bucket.get_all_keys())
    bucket.delete()
    print('{0} is deleted'.format(bucket.name))



"""
Serve this site locally.
"""

import sys
import SimpleHTTPServer
import SocketServer


class MyTCPServer(SocketServer.TCPServer):
    """
    Release the socket when killed.
    """
    allow_reuse_address = True


def local(arguments):
    """
    Runs the site in a local server on port 8000 or specified port.

    Usage: widely local [-p <PORT> | --port <PORT>]
    """
    host = '0.0.0.0'
    port = arguments['<PORT>']
    if port:
        try:
            port = int(port)
        except ValueError:
            print('<PORT> must be an integer.')
            sys.exit(1)
    else:
        port = 8000

    Handler = SimpleHTTPServer.SimpleHTTPRequestHandler
    httpd = MyTCPServer((host, port), Handler)
    url = 'http://' + host + ':' + str(port)
    print("serving at " + url)
    import webbrowser

    webbrowser.open_new_tab(url)
    httpd.serve_forever()



import os

from setuptools import setup, find_packages
from pip.req import parse_requirements


basedir = os.path.dirname(__file__)
requirements_path = os.path.join(basedir, 'requirements.txt')

# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_requirements = parse_requirements(
    requirements_path, session=False
)

# Convert to setup's list of strings format:
requirements = [str(ir.req) for ir in install_requirements]

setup(
    name='widely',
    version='0.16',
    description='Static Site as a Service using AWS S3',
    long_description=open('README.rst').read(),
    author='Kyle Marek-Spartz and Michael Burling',
    author_email='kyle.marek.spartz@gmail.com',
    py_modules=['widely'],
    url='http://www.celador.mn/widely',
    include_package_data=True,
    packages=find_packages(exclude=['tests*']),
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'widely = widely:main',
        ],
    },
    test_suite='nose.collector',
    classifiers=["Development Status :: 3 - Alpha",
                 "Environment :: Console",
                 "Intended Audience :: Developers",
                 "Intended Audience :: Information Technology",
                 "Intended Audience :: System Administrators",
                 "License :: OSI Approved :: MIT License",
                 "Programming Language :: Python :: 2.7",
                 "Programming Language :: Python :: 2 :: Only",
                 "Topic :: Internet",
                 "Topic :: Internet :: WWW/HTTP",
                 "Topic :: Internet :: WWW/HTTP :: Site Management",
                 "Topic :: Utilities"],
    license='MIT'
)



"""
Opens a bucket in the webbrowser.
"""

from __future__ import print_function
import webbrowser

from widely.bucket import get_current_or_specified_bucket


def _open(arguments):
    """
    Loads the running specified or current site in the webbrowser.

    Usage: widely open [--site <SITENAME>]
    """
    bucket = get_current_or_specified_bucket(arguments)
    url = 'http://' + bucket.get_website_endpoint()
    print('Opening {0}...'.format(bucket.name), end="")
    webbrowser.open_new_tab(url)
    print('done')



"""
Diffs are used to compare and reconcile local and remote data.
"""

from glob import glob
import hashlib
import os
from prettytable import PrettyTable


class Diff(object):
    """
    Diff class. Diff is used to track changes between current
    directory and AWS S3.
    """
    NotRemote = 'NotRemote'
    NotLocal = 'NotLocal'
    Modified = 'Modified'


def generate_diffs(bucket):
    """
    Returns a list of diffs or changes between current directory and AWS S3.
    """
    # TODO: Use difflib?
    # TODO: Don't use MD5!
    ignored = set(['.widely', '.widelyignore'])

    try:
        with open('.widelyignore', 'r') as f:
            _ignored = [glob(line) for line in f.read().splitlines()]
            specified_ignored = set(
                item
                for sublist in _ignored
                for item in sublist
            )
            ignored = ignored | specified_ignored
    except IOError:
        pass

    def get_local_keys():
        """
        Walk the current directory, recursively yielding the paths and
        hashes for each non-ignored file.
        """
        for root, _, files in os.walk(os.curdir):
            for _file in files:
                path = os.path.join(root, _file)[2:]
                if path not in ignored:
                    with open(path, 'r') as f:
                        yield path, '"' + hashlib.md5(
                            f.read()).hexdigest() + '"'

    def get_remote_keys():
        """
        Gets the list of keys and hashes from the bucket.
        """
        remote_keys = bucket.get_all_keys()
        for key in remote_keys:
            yield key.name, key.etag

    local_keys = dict(get_local_keys())
    remote_keys = dict(get_remote_keys())

    diffs = list()

    for key in local_keys:
        if os.path.isdir(key):
            continue
        if key in remote_keys:
            # The file is in both places
            if local_keys[key] == remote_keys[key]:
                continue
            else:
                diffs.append((Diff.Modified, key))
        else:
            # The file is local but not remote.
            diffs.append((Diff.NotRemote, key))
    for key in remote_keys:
        if key not in local_keys:
            # The file is remote but not local.
            diffs.append((Diff.NotLocal, key))

    return diffs


def show_diffs(diffs):
    """
    Prints a table of diffs.
    """
    print('This would make the following changes:')
    table = PrettyTable(['Diff', 'Key'])
    sorted_diffs = sorted(diffs, key=lambda (_, snd): snd)
    for diff, key in sorted_diffs:
        table.add_row([diff, key])
    print(table)


def run_diffs(diffs, bucket, local_changes=None):
    """
    Takes diffs and makes the necessary changes, either pushing or
    pulling files.
    """
    if type(local_changes) is not bool:
        raise ValueError('local_changes must be set to a bool')

    if local_changes:
        for diff, key in diffs:
            if diff == Diff.NotRemote:
                # Delete local file
                print('removing (local): {0}'.format(key))
                os.remove(key)
            elif diff == Diff.NotLocal or diff == Diff.Modified:
                # Pull remote file
                print('pulling: {0}'.format(key))
                with open(key, 'w') as f:
                    bucket.get_key(key).get_contents_to_file(f)
    else:
        # Remote changes
        for diff, key in diffs:
            if diff == Diff.NotRemote or diff == Diff.Modified:
                # Push local file
                print('pushing: {0}'.format(key))
                with open(key, 'r') as f:
                    k = bucket.new_key(key)
                    k.set_contents_from_file(f)
                    k.make_public()
            elif diff == Diff.NotLocal:
                # Delete remote file
                print('removing (remote): {0}'.format(key))
                bucket.get_key(key).delete()



"""
Manage and display help messages.
"""

import widely
from widely.commands.push import push
from widely.commands.pull import pull
from widely.commands.status import status
from widely.commands.sites import sites, sites_create, sites_info, \
    sites_delete, sites_rename, sites_copy
from widely.commands.local import local
from widely.commands.logs import logs
from widely.commands.version import version
from widely.commands.auth import auth_login, auth_logout, auth_whoami
from widely.commands.domains import domains
from widely.commands.open import _open


def _help(arguments):
    """
    Look up help messages.

    Usage:
    widely help <TOPIC>

    where <TOPIC> is one of the following:

    """
    help_messages = {
        'version': version.__doc__,
        'help': _help.__doc__,
        'auth:login': auth_login.__doc__,
        'login': auth_login.__doc__,
        'auth:logout': auth_logout.__doc__,
        'logout': auth_logout.__doc__,
        'auth:whoami': auth_whoami.__doc__,
        'domains': domains.__doc__,
        'local': local.__doc__,
        'logs': logs.__doc__,
        'open': _open.__doc__,
        'sites': sites.__doc__,
        'sites:create': sites_create.__doc__,
        'sites:info': sites_info.__doc__,
        'sites:delete': sites_delete.__doc__,
        'sites:copy': sites_copy.__doc__,
        'sites:rename': sites_rename.__doc__,
        'status': status.__doc__,
        'push': push.__doc__,
        'pull': pull.__doc__
    }
    topic = arguments['<TOPIC>']
    if topic and topic in help_messages:
        message = 'help about {0}\n'.format(topic)
        message += help_messages[topic]
        if topic == 'help':
            message += '\n    '.join(help_messages.keys())
        print(message)
    else:
        print(widely.__doc__.strip("\n"))
